//      3Cubes.cpp      
// Copyright Tom Hubin 2 April 2003
// Email: thubin@earthlink.net
// Phone: 301-776-7063, Laurel MD 20708

// compiler header files

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <math.h>
#include <mem.h>

// useful procedures

#define BELL 7
#define BEEP putch(BELL)
#define max(a,b) (((a)>(b))?(a):(b))
#define min(a,b) (((a)<(b))?(a):(b))

#define FALSE 0
#define TRUE  1

// useful constants

const double meter = 1.0 ;               // (meter) unit of distance
const double millimeter = meter * 1e-3 ; // (m) multiplier from millimeters to meters
const double micron = meter * 1e-6 ;     // (m) multiplier from microns to meters
const double inch = 25.4 * millimeter ;  // (m) multiplier from inches to meters

const double second = 1.0 ;              // (second) unit of time
const double minute = 60 * second ;      // (sec) multiplier for time
const double hour = 60 * minute ;        // (sec) multiplier for time
const int  cw = -1 ;                     // circle cut direction
const int ccw = +1 ;

// structures

typedef struct
{
    double cubeWidth ;          // (m) cube width

    double vrr ;                // (m^3/sec) volume removal rate
    double xyMaxSpeed ;         // (m/sec) max horizontal feedrate
    double zSpeed ;             // (m/sec) vertical feedrate
    double diamMin ;            // (m) minimum bore diameter
    double diamMax ;            // (m) maximum bore diameter
    double diamStepSize ;       // (m) bore diameter step size

    double centerX ;            // (m) x center of bore
    double centerY ;            // (m) y center of bore
    double topZ ;               // (m) z top of bore
    double bottomZ ;            // (m) z bottom of bore
    double tallestZ ;           // (m) z tallest obstacle to avoid
    double depthStepSize ;      // (m) bore depth step size

    double endmillDiam ;        // (m) endmill diameter
} CUBE ;

//      local prototypes

static int GenerateBoreGCode ( FILE* outFile, double lengthUnit, double speedUnit, CUBE* cube ) ;

int main ( void )
{
    int rtc = 0 ;
    FILE* outFile = NULL ;
    char* outFileName = "3Cubes.cnc" ;       // G code output file name
    const int nCubes = 3 ;      // number of cubes
    CUBE cube[nCubes] ;         // data for cubes
    double lengthUnit = inch ;
    double speedUnit = lengthUnit/minute ;

    outFile = fopen ( outFileName, "w" ) ;
    if ( outFile == NULL )
    {
        BEEP ;
        printf ( "Cannot open file [%s] for output\n", outFileName ) ;
        printf ( "Hit a key to terminate\n" ) ;
        getch() ;
        rtc = 1 ;
        goto quit ;
    }

    memset ( &cube[0], 0, sizeof(cube[0]) ) ;
    cube[0].cubeWidth = 1.950 * inch ;    // (m) outermost cube width

    fprintf ( outFile, "M00 ; 3Cubes bore for %f inch cube\n",
        cube[0].cubeWidth/lengthUnit ) ;
    fprintf ( outFile, "M00 ; Copyright Tom Hubin 22 July 2003\n" ) ;
    fprintf ( outFile, "M00 ; Email: thubin@earthlink.net\n" ) ;
    fprintf ( outFile, "M00 ; Phone: 301-776-7063, Laurel MD 20708\n" ) ;
    fprintf ( outFile, "\n" ) ;

    fprintf ( outFile, "G90 ; absolute coordinates\n" ) ;

    if ( lengthUnit == millimeter )
        fprintf ( outFile, "G71 ; millimeters\n" ) ;
    else
        fprintf ( outFile, "G70 ; inches\n" ) ;

        // create Gcode for first bore

    memcpy ( &cube[1], &cube[0], sizeof(cube[1]) ) ;
    cube[1].cubeWidth = 1.0 * inch ;
    cube[1].endmillDiam = 3.0/8.0 * inch ;    // (m) endmill diameter
    cube[1].centerX = 0.0 * millimeter ;      // (m) x center of bore
    cube[1].centerY = 0.0 * millimeter ;      // (m) y center of bore
    cube[1].topZ = 0.0 * millimeter ;         // (m) z top of bore
    cube[1].tallestZ = 0.0 * millimeter ;     // (m) z tallest obstacle to avoid

    cube[1].vrr = 3.0 * inch*inch*inch/hour ;
    cube[1].xyMaxSpeed = 14 * inch/minute ;
    cube[1].zSpeed = 0.5 * inch/minute ;
    cube[1].diamStepSize = 0.020 * inch ;
    cube[1].depthStepSize = 0.75 * inch ;
    cube[1].diamMin = cube[1].endmillDiam ;
    cube[1].diamMax = cube[1].cubeWidth * sqrt(2.0) - 0.020 * inch ;
    cube[1].bottomZ = -0.5 * ( cube[0].cubeWidth - cube[1].cubeWidth ) ;

        // create Gcode for second bore

    memcpy ( &cube[2], &cube[1], sizeof (cube[2]) ) ;
    cube[2].cubeWidth = 0.5 * inch ;
    cube[2].diamStepSize = 0.020 * inch ;
    cube[2].depthStepSize = 0.75 * inch ;
    cube[2].diamMin = cube[2].endmillDiam ;
    cube[2].diamMax = cube[2].cubeWidth * sqrt(2.0) - 0.020 * inch ;
    cube[2].topZ = cube[1].bottomZ ;
    cube[2].bottomZ = -0.5 * ( cube[0].cubeWidth - cube[2].cubeWidth ) ;

    fprintf ( outFile, "M00 ; Top center of bore is at (%+0.6lf,%+0.6lf,%+0.6lf)\n",
        cube[1].centerX/lengthUnit,
        cube[1].centerY/lengthUnit,
        cube[1].topZ/lengthUnit ) ;
    fprintf ( outFile, "M00 ; Ref (right,rear,top) = (%+0.6lf,%+0.6lf,%+0.6lf)\n",
        cube[0].cubeWidth/lengthUnit/2,
        cube[0].cubeWidth/lengthUnit/2,
        cube[1].topZ/lengthUnit ) ;

    int iCube ;
    for ( iCube = 1 ; iCube < nCubes ; iCube++ )
    {
        if ( cube[iCube].endmillDiam != cube[iCube-1].endmillDiam )
        {
            fprintf ( outFile, "M00 ; Load %+0.6lf endmill and touch surface\n",
                cube[iCube].endmillDiam/lengthUnit ) ;
            fprintf ( outFile, "G92 Z %0.6lf ; Define Z height\n",
                cube[1].topZ/lengthUnit ) ;

            fprintf ( outFile, "\n    ; Position to start bore\n" ) ;
            fprintf ( outFile, "G00 Z %+0.6lf ; Over the tallest point\n",
                ( cube[iCube].tallestZ + 0.050 * inch )/lengthUnit ) ;
            fprintf ( outFile, "M00 ; Tighten bit and start spindle 2800 RPM\n" ) ;

            fprintf ( outFile, "G00 X %+0.6lf Y %+0.6lf ; Over bore center\n",
                cube[iCube].centerX/lengthUnit,
                cube[iCube].centerY/lengthUnit ) ;
        }
        rtc = GenerateBoreGCode ( outFile, lengthUnit, speedUnit, &cube[iCube] ) ;
        if ( rtc )
        {
            goto quit ;
        }
    }

        // retract bit to top of cube

    fprintf ( outFile, "\n    ; retract above top of cube\n" ) ;
    fprintf ( outFile, "G00 Z %+0.6lf\n",
        ( cube[0].tallestZ + 0.050 * inch ) /lengthUnit ) ;

    fprintf ( outFile, "\nM02 ; end of program\n" ) ;

    fclose ( outFile ) ;
    outFile = NULL ;

quit:
    if ( outFile )
    {
        fclose ( outFile ) ;
        outFile = NULL ;
    }

    return (rtc) ;
}

int GenerateBoreGCode ( FILE* outFile, double lengthUnit, double speedUnit, CUBE* srcCube )
{
    int rtc = 0 ;
    CUBE bore ;

    memcpy ( &bore, srcCube, sizeof(bore) ) ;

        // avoid zero circle radius error

    bore.diamMin = max ( bore.diamMin, bore.endmillDiam + 0.001 * inch ) ;

        // compute loop counts

    int boreDiamNSteps = ceil( fabs( bore.diamMax - bore.diamMin ) / bore.diamStepSize ) ;
    double boreDiamStepSize = ( bore.diamMax - bore.diamMin ) / boreDiamNSteps ;
    int boreDepthNSteps = ceil ( fabs ( bore.topZ - bore.bottomZ ) / bore.depthStepSize ) ;
    double boreDepthStepSize = ( bore.topZ - bore.bottomZ ) / boreDepthNSteps ;

    fprintf ( outFile, "\nM00 ; boreDiamNSteps = %i    boreDepthNSteps = %i\n",
        boreDiamNSteps, boreDepthNSteps ) ;

    int iDepth, iDiam ;
    for ( iDepth = 1 ; iDepth <= boreDepthNSteps ; iDepth++ )
    {
        double boreFinishX, boreOffsetR2 ;

        double boreDepth = bore.topZ - ( bore.topZ - bore.bottomZ ) * iDepth / boreDepthNSteps ;

        fprintf ( outFile, "\n    ; center the bit at depth %+0.6lf\n",
            boreDepth/lengthUnit ) ;
        fprintf ( outFile, "G01 X %+0.6lf Y %+0.6lf F%0.3lf\n",
            bore.centerX/lengthUnit,
            bore.centerY/lengthUnit,
            bore.xyMaxSpeed/speedUnit ) ;
        fprintf ( outFile, "G01 Z %+0.6lf F%0.3lf\n",
            boreDepth/lengthUnit,
            bore.zSpeed/speedUnit ) ;

        for ( iDiam = 1 ; iDiam <= boreDiamNSteps ; iDiam++ )
        {
            double boreDiam2 = bore.diamMin + ( bore.diamMax - bore.diamMin ) * iDiam / boreDiamNSteps ;
            double boreDiam1 = boreDiam2 - ( bore.diamMax - bore.diamMin ) * 0.5 / boreDiamNSteps ;

            boreOffsetR2 = ( boreDiam2 - bore.endmillDiam ) / 2 ;
            double boreOffsetR1 = ( boreDiam1 - bore.endmillDiam ) / 2 ;

            fprintf ( outFile, "\n    ; cut %0.6lf diam nominal circle\n",
                boreDiam2/lengthUnit ) ;

            boreFinishX = bore.centerX - boreOffsetR2 ;
            double boreStartX  = bore.centerX + boreOffsetR2 ;

                // feedrate for circle
            double D2 = boreDiam2 ;
            double A2 = D2 * D2 * M_PI / 4.0 ;

            double D1 = boreDiam2 - boreDiamStepSize ;
            double A1 = D1 * D1 * M_PI / 4.0 ;

            double vr = ( A2 - A1 ) * boreDepthStepSize ;
            double time = vr / bore.vrr ;
            double pathLength = 2 * boreOffsetR2 * M_PI ;

            double circleSpeed = pathLength / time ;

            fprintf ( outFile, "G02 X %+0.6lf Y %+0.6lf I %+0.6lf J 0.0 F%0.3lf\n",
                boreStartX/lengthUnit,
                bore.centerY/lengthUnit,
                boreOffsetR1/lengthUnit,
                circleSpeed/speedUnit ) ;

            fprintf ( outFile, "G02 X %+0.6lf Y %+0.6lf I %+0.6lf J 0.0 F%0.3lf\n",
                boreFinishX/lengthUnit,
                bore.centerY/lengthUnit,
                -boreOffsetR2/lengthUnit,
                circleSpeed/speedUnit ) ;

            if ( iDiam == boreDiamNSteps )
            {
                fprintf ( outFile, "G02 X %+0.6lf Y %+0.6lf I %+0.6lf J 0.0 F%0.3lf\n",
                    boreFinishX/lengthUnit,
                    bore.centerY/lengthUnit,
                    boreOffsetR2/lengthUnit,
                    circleSpeed/speedUnit ) ;

                fprintf ( outFile, "G03 X %+0.6lf Y %+0.6lf I %+0.6lf J 0.0 F%0.3lf\n",
                    boreFinishX/lengthUnit,
                    bore.centerY/lengthUnit,
                    boreOffsetR2/lengthUnit,
                    circleSpeed/speedUnit ) ;
            }
        }

            // if bore complete then
            // cleanup wall with climb helix

        if ( iDepth == boreDepthNSteps )
        {
            fprintf ( outFile, "\n" ) ;
            fprintf ( outFile, "    ; retract in fast fine helices\n" ) ;
            fprintf ( outFile, "\n" ) ;

            double zRetractStep = 0.001 * inch ;
            for ( int done=FALSE ; !done ; zRetractStep *= 2 )
            {
                double zNext = bore.bottomZ + zRetractStep ;
                if ( zNext >= ( bore.topZ + 0.010 * inch ) )
                {
                    zNext = bore.topZ + 0.010 * inch ;
                    done = TRUE ;
                }

                fprintf ( outFile, "G73 Z %+0.6lf X %+0.6lf Y %+0.6lf I %+0.6lf F%0.3lf J 0.0\n",
                    zNext/lengthUnit, 
                    boreFinishX/lengthUnit,
                    bore.centerY/lengthUnit,
                    boreOffsetR2/lengthUnit,
                    bore.xyMaxSpeed/speedUnit ) ;
            }
        }

            // retract bit to center of bore

        fprintf ( outFile, "\n    ; retract to just above center of finished bore\n" ) ;

        double zRetract = boreDepth + 0.010 * inch ;
        double xRetract = bore.centerX ;
        double yRetract = bore.centerY ;

        fprintf ( outFile, "G01 X %+0.6lf Y %+0.6lf Z %+0.6lf F%0.3lf\n",
            xRetract/lengthUnit,
            yRetract/lengthUnit,
            zRetract/lengthUnit,
            bore.xyMaxSpeed/speedUnit ) ;

        fprintf ( outFile, "G01 X %+0.6lf Y %+0.6lf F%0.3lf\n",
            bore.centerX/lengthUnit,
            bore.centerY/lengthUnit,
            bore.xyMaxSpeed/speedUnit ) ;
    }

quit:

    return (rtc) ;
}

